import logging
from distutils.version import LooseVersion
from pathlib import Path

import pandas as pd
import geopandas as gpd
import pyproj
import rasterio
from shapely.wkt import loads
from shapely.geometry import Point
from shapely.geometry.base import BaseGeometry
from fiona._err import CPLE_OpenFailedError
from fiona.errors import DriverError

from ..config.logging_config import logger

logger = logging.getLogger(__name__)

"""Utility funtions adapted from Solaris: https://github.com/CosmiQ/solaris/blob/main/solaris/utils/core.py"""

def rasterio_load(im):
    """Load a rasterio image from a path or a rasterio object.
    Args:
        im (str or Path or rasterio.DatasetReader): Path to the image or a rasterio object. 
        
    Returns:
        rasterio.DatasetReader: rasterio dataset.
    """
    if isinstance(im, (str, Path)):
        return rasterio.open(im)
    elif isinstance(im, rasterio.DatasetReader):
        return im
    else:
        raise ValueError("{} is not an accepted image format for rasterio.".format(im))

def gdf_load(gdf):
    """Load a GeoDataFrame from a path or a GeoDataFrame object.

    Args:
        gdf (str or Path or GeoDataFrame): Path to the GeoDataFrame or a GeoDataFrame object.

    Returns:
        GeoDataFrame: GeoDataFrame.
    """
    if isinstance(gdf, (str, Path)):
        try:
            return gpd.read_file(gdf)
        except (DriverError, CPLE_OpenFailedError):
            logger.warning(f"GeoDataFrame couldn't be loaded: either {gdf} isn't a valid"
                           f" path or it isn't a valid vector file. Returning an empty"
                           f" GeoDataFrame.")
            return gpd.GeoDataFrame()
    elif isinstance(gdf, gpd.GeoDataFrame):
        return gdf
    else:
        raise ValueError(f"{gdf} is not an accepted GeoDataFrame format.")

def df_load(df):
    """Check if `df` is already loaded in, if not, load from file."""
    if isinstance(df, str):
        if df.lower().endswith('json'):
            return gdf_load(df)
        else:
            return pd.read_csv(df)
    elif isinstance(df, pd.DataFrame):
        return df
    else:
        raise ValueError(f"{df} is not an accepted DataFrame format.")

def check_geom(geom):
    """Check if a geometry is loaded in.

    Returns the geometry if it's a shapely geometry object. If it's a wkt
    string or a list of coordinates, convert to a shapely geometry.
    """
    if isinstance(geom, BaseGeometry):
        return geom
    elif isinstance(geom, str):  # assume it's a wkt
        return loads(geom)
    elif isinstance(geom, list) and len(geom) == 2:  # coordinates
        return Point(geom)

def check_crs(input_crs, return_rasterio=False):
    """Check the input CRS and return a pyproj CRS object.

    Args:
        input_crs (str or pyproj.CRS): Input CRS.
        return_rasterio (bool, optional): Returns rasterio crs object. Defaults to False.

    Returns:
        rasterio or pyproj crs: CRS object.
    """
    if not isinstance(input_crs, pyproj.CRS) and input_crs is not None:
        out_crs = pyproj.CRS(input_crs)
    else:
        out_crs = input_crs

    if return_rasterio:
        if LooseVersion(rasterio.__gdal_version__) >= LooseVersion("3.0.0"):
            out_crs = rasterio.crs.CRS.from_wkt(out_crs.to_wkt())
        else:
            out_crs = rasterio.crs.CRS.from_wkt(out_crs.to_wkt("WKT1_GDAL"))

    return out_crs


import logging
import os
import sys
from typing import Any, Dict, Iterator, Optional, Tuple, Union, cast

import numpy as np
import rasterio as rio
import scipy.signal.windows as w
import torch
from rasterio.crs import CRS
from scipy.special import expit
from rasterio.windows import Window
from torch import Tensor
from torch.nn import functional as F
from torchgeo.datasets import GeoDataset
from torchgeo.datasets.utils import BoundingBox
from torchgeo.samplers import GeoSampler
from torchgeo.samplers.utils import _to_tuple, tile_to_chips

from .config.logging_config import logger
from .utils.helpers import validate_asset_type

logger = logging.getLogger(__name__)


class RasterDataset(GeoDataset):
    """
    A dataset class for raster data.
    
    Attributes:
        image_asset: The path to the image asset.
        src: The rasterio dataset object.
        cmap: The colormap of the image asset.
        crs: The coordinate reference system of the image asset.
        res: The resolution of the image asset.
        bands: The number of bands in the image asset.
        index: The rtree index of the image asset.
    """
    def __init__(self, image_asset: str, bbox: str = None) -> None:
        """Initializes a RasterDataset object.
        
        Args:
            image_asset (str): The path to the image asset.
            bounding_box (str): The bounding box of the image asset.
        """
        super().__init__()
        self.image_asset = validate_asset_type(image_asset)
        
        if self.image_asset is None:
            logger.error(f"image asset is neither a tiff image | tiff url")
            raise ValueError("Invalid image_asset type")
        
        with rio.open(self.image_asset) as src:
            self.src = src
            try:
                self.cmap = src.colormap(1)
            except ValueError:
                pass
            
            crs = src.crs
            res = src.res[0]
            bands = src.count
            minx, miny, maxx, maxy = src.bounds
        
            mint: float = 0
            maxt: float = sys.maxsize
            
            if bbox is None:
                bbox = BoundingBox(minx=minx, 
                                   miny=miny, 
                                   maxx=maxx, 
                                   maxy=maxy, 
                                   mint=mint, maxt=maxt)
            else:
                bbox = tuple(map(float, bbox.split(', ')))
                bbox = BoundingBox(minx=bbox[0], 
                                   miny=bbox[1], 
                                   maxx=bbox[2], 
                                   maxy=bbox[3], 
                                   mint=mint, maxt=maxt)
                
            coords = (minx, maxx, miny, maxy, mint, maxt)
            self.index.insert(0, coords, self.image_asset)
        
        self._crs = cast(CRS, crs)
        self.res = cast(float, res)
        self.bands = bands
        self.bbox = bbox
    
    def __getitem__(self, query: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get a sample from the dataset.

        Args:
            query: A dictionary containing the query parameters.

        Returns:
            A dictionary containing the sample data.
        """
        filepath = query['path']
        window = query["window"]
        pixel_coords = query["pixel_coords"]
        patch_size = pixel_coords[-1]
        
        data = self._get_tensor(pixel_coords, patch_size)
        sample = {"image": data, 
                  "crs": self.crs,
                  "pixel_coords": pixel_coords, 
                  "window": window,
                  "path": filepath}

        return sample
    
    def _get_tensor(self, query, size):
        """
        Get a patch based on the given query (pixel coordinates).

        Args:
            query: The pixel coordinates of the patch.
            size: The desired patch size.

        Returns:
            A torch tensor patch.
        """
        (x_min, y_min, patch_width, patch_height) = query
        
        window = Window.from_slices(slice(y_min, y_min + patch_height),
                                    slice(x_min, x_min + patch_width))
        
        with rio.open(self.image_asset) as src:
            dest = src.read(window=window)
            if dest.dtype == np.uint16:
                dest = dest.astype(np.int32)
            elif dest.dtype == np.uint32:
                dest = dest.astype(np.int64)

            tensor = torch.tensor(dest)
            tensor = self.pad_patch(tensor, size)
        
        return tensor
    
    @staticmethod
    def pad_patch(x: Tensor, patch_size: int):
        """
        Pad the patch to desired patch_size.

        Args:
            x: The tensor patch to pad.
            patch_size: The desired patch size.

        Returns:
            The padded tensor patch.
        """
        h, w = x.shape[-2:]
        pad_h = patch_size - h
        pad_w = patch_size - w
        # pads are described starting from the last dimension and moving forward.
        x = F.pad(x, (0, pad_w, 0, pad_h))
        return x


class InferenceSampler(GeoSampler):
    """Class for creating an inference sampler.

    This class extends GeoSampler and is designed for generating patches
    for inference on a GeoDataset.

    Attributes:
        dataset (GeoDataset): The dataset to generate patches from.
        size (Union[Tuple[float, float], float]): Dimensions of each patch.
        stride (Union[Tuple[float, float], float]): Distance to skip between each patch.
        roi (Optional[BoundingBox]): Region of interest to sample from.
    """
    def __init__(self,
                 dataset: GeoDataset,
                 size: Union[Tuple[float, float], float],
                 stride: Union[Tuple[float, float], float],
                 roi: Optional[BoundingBox] = None,
                 ) -> None:
        """
        Initializes an InferenceSampler object.

        Args:
            dataset (GeoDataset): A GeoDataset object.
            size (Union[Tuple[float, float], float]): The size of the patch.
            stride (Union[Tuple[float, float], float]): The stride of the patch.
            roi (Optional[BoundingBox], optional): A BoundingBox object. Defaults to None.
        """
        super().__init__(dataset, roi)
        self.size = _to_tuple(size)
        self.patch_size = self.size
        self.stride = _to_tuple(stride)
        
        # Generates 9 2D signal windows of patch size that covers edge and corner coordinates
        self.windows = torch.tensor(self.generate_corner_windows(self.patch_size[0]), dtype=torch.float32)
        self.size_in_crs_units = (self.size[0] * self.res, self.size[1] * self.res)
        self.stride_in_crs_units = (self.stride[0] * self.res, self.stride[1] * self.res)
        self.hits = []
        self.hits_small = []
        for hit in self.index.intersection(tuple(self.roi), objects=True):
            bounds = BoundingBox(*hit.bounds)
            if (bounds.maxx - bounds.minx >= self.size_in_crs_units[1]
                and bounds.maxy - bounds.miny >= self.size_in_crs_units[0]):
                self.hits.append(hit)
            else:
                self.hits_small.append(hit)
        
        self.length = 0
        for hit in self.hits:
            bounds = BoundingBox(*hit.bounds)
            rows, cols = tile_to_chips(bounds=bounds, size=self.size_in_crs_units, stride=self.stride_in_crs_units)
            self.length += rows * cols
        
        for hit in self.hits_small:
            bounds = BoundingBox(*hit.bounds)
            self.length += 1
        
        for hit in self.hits + self.hits_small:
            if hit in self.hits:
                bounds = BoundingBox(*hit.bounds)
                self.im_height = round((bounds.maxy - bounds.miny) / self.res)
                self.im_width = round((bounds.maxx - bounds.minx) / self.res)
            else:
                bounds = BoundingBox(*hit.bounds) 
                self.im_height = round((bounds.maxy - bounds.miny) / self.res)
                self.im_width = round((bounds.maxx - bounds.minx) / self.res)
                
    def __iter__(self) -> Iterator[Dict[str, Any]]:
        """
        Yields a dictionary containing the pixel coordinates, path, and window.
        """
        for hit in self.hits + self.hits_small:
            if hit in self.hits:
                y_steps = int(np.ceil((self.im_height - self.size[0]) / self.stride[0]) + 1)
                x_steps = int(np.ceil((self.im_width - self.size[1]) / self.stride[1]) + 1)
                for y in range(y_steps):
                    if self.stride[0] * y + self.size[0] > self.im_height:
                        y_min = self.im_height - self.size[0]
                    else:
                        y_min = self.stride[0] * y
                    for x in range(x_steps):
                        if self.stride[1] * x + self.size[1] > self.im_width:
                            x_min = self.im_width - self.size[1]
                        else:
                            x_min = self.stride[1] * x
                        # get center, border and corner windows
                        border_x, border_y = 1, 1
                        if y == 0: border_x = 0
                        if x == 0: border_y = 0
                        if y == y_steps - 1: border_x = 2
                        if x == x_steps - 1: border_y = 2
                        # Select the right window
                        current_window = self.windows[border_x, border_y]
                        query = {"pixel_coords": (x_min, y_min, self.patch_size[1], self.patch_size[0]),
                                 "path": cast(str, hit.object),
                                 "window": current_window}
                        yield query
            else:
                x_min, y_min = (0, 0)
                current_window = torch.ones((self.patch_size[0], self.patch_size[1]))
                query = {"pixel_coords": (x_min, y_min, self.patch_size[1], self.patch_size[0]),
                         "path": cast(str, hit.object),
                         "window": current_window}
                yield query
    
    def __len__(self) -> int:
        """
        Returns the number of samples over the ROI.

        Returns:
            int: The number of patches that will be sampled.
        """
        return self.length
    
    @staticmethod
    def generate_corner_windows(window_size: int) -> np.ndarray:
        """
        Generates 9 2D signal windows that covers edge and corner coordinates
        
        Args:
            window_size (int): The size of the window.

        Returns:
            np.ndarray: 9 2D signal windows stacked in array (3, 3).
        """
        step = window_size >> 1
        window = np.matrix(w.hann(M=window_size, sym=False))
        window = window.T.dot(window)
        window_u = np.vstack([np.tile(window[step:step+1, :], (step, 1)), window[step:, :]])
        window_b = np.vstack([window[:step, :], np.tile(window[step:step+1, :], (step, 1))])
        window_l = np.hstack([np.tile(window[:, step:step+1], (1, step)), window[:, step:]])
        window_r = np.hstack([window[:, :step], np.tile(window[:, step:step+1], (1, step))])
        window_ul = np.block([[np.ones((step, step)), window_u[:step, step:]],
                              [window_l[step:, :step], window_l[step:, step:]]])
        window_ur = np.block([[window_u[:step, :step], np.ones((step, step))],
                              [window_r[step:, :step], window_r[step:, step:]]])
        window_bl = np.block([[window_l[:step, :step], window_l[:step, step:]],
                              [np.ones((step, step)), window_b[step:, step:]]])
        window_br = np.block([[window_r[:step, :step], window_r[:step, step:]],
                              [window_b[step:, :step], np.ones((step, step))]])
        return np.array([[window_ul, window_u, window_ur],
                         [window_l, window, window_r],
                         [window_bl, window_b, window_br]])
    

class InferenceMerge:
    """
    A class for merging inference results.
    
    Attributes:
        height (int): The padded height of roi.
        width (int): The padded width of roi.
        classes (int): The number of classes.
        device (torch.device): The device to use for computation.
        image (np.ndarray): The merged image.
        norm_mask (np.ndarray): The normalization mask.
    
    """
    def __init__(self, 
                 height: int, 
                 width: int, 
                 classes: int, 
                 device: torch.device) -> None:
        """
        Initializes a new instance of the InferenceMerge class.

        Args:
            height (int): The padded height of roi.
            width (int): The padded width of roi.
            classes (int): The number of classes.
            device (torch.device): The device to use for computation.
        """
        self.height = height
        self.width = width
        self.classes = classes
        self.device = device
        self.image = np.zeros((self.classes, self.height, self.width), dtype=np.float16)
        self.norm_mask = np.ones((1, self.height, self.width), dtype=np.float16)
        # self.image = torch.zeros((self.classes, self.height, self.width), dtype=torch.float16, device=self.device)
        # self.norm_mask = torch.ones((1, self.height, self.width), dtype=torch.float16, device=self.device)
    
    @torch.no_grad()
    def merge_on_cpu(self, batch: torch.Tensor, windows: torch.Tensor, pixel_coords):
        """
        Merge the patches on CPU.

        Args:
            batch (torch.Tensor): The batch of inference results.
            windows (torch.Tensor): The windows used for inference.
            pixel_coords (list): The pixel coordinates of the patches.

        Returns:
            None
        """
        for output, window, (x, y, patch_width, patch_height) in zip(batch, windows, pixel_coords):
            if self.classes == 1:
                output = F.sigmoid(output) * window
            else:
                output = F.softmax(output, dim=0) * window
            self.image[:, y : y + patch_height, x : x + patch_width] += output.cpu().numpy()
            self.norm_mask[:, y : y + patch_height, x : x + patch_width] += window.cpu().numpy()
    
    @torch.no_grad()
    def merge_on_gpu(self,):
        """
        Merge the patches on GPU.
        """
        pass

    def save_as_tiff(self, 
                     height: int, 
                     width: int, 
                     output_meta: dict, 
                     output_path: os.PathLike) -> torch.Tensor:
        """
        Save mask to file. 

        Args:
            height (int): The height of the output mask.
            width (int): The width of the output mask.
            output_meta (dict): The meta data of the output mask.
            output_path (os.PathLike): The path to save the output mask.

        Returns:
            None
        """
        threshold = 0.5
        self.image /= self.norm_mask
        
        # Binary mask
        if self.image.shape[0] == 1:
            self.image = expit(self.image)
            self.image = np.where(self.image > threshold, 1, 0).astype(np.uint8)
        else:
            self.image = np.argmax(self.image, axis=0).astype(np.uint8)
            # self.image = torch.argmax(self.image, dim=0).byte().cpu().numpy()
            
        self.image = self.image[np.newaxis, :height, :width]
        output_meta.update({"driver": "GTiff",
                            "height": self.image.shape[1],
                            "width": self.image.shape[2],
                            "count": self.image.shape[0],
                            "dtype": 'uint8',
                            "compress": 'lzw'})
        with rio.open(output_path, 'w+', **output_meta) as dest:
            dest.write(self.image)
        logger.info(f"Mask saved to {output_path}")


"""Adapted from Solaris: https://github.com/CosmiQ/solaris/tree/main/solaris """

import logging
import os

import numpy as np
import geopandas as gpd
import pandas as pd
import rasterio
import shapely
import json

from shapely.wkt import loads
from affine import Affine
from rasterio import features
from rasterio.warp import transform_bounds
from rtree.core import RTreeError
from shapely.geometry import Polygon, mapping, shape

from ..config.logging_config import logger
from .geo import check_crs, check_geom, df_load, gdf_load, rasterio_load

logger = logging.getLogger(__name__)


def _reduce_geom_precision(geom, precision=2):
    geojson = mapping(geom)
    geojson['coordinates'] = np.round(np.array(geojson['coordinates']),
                                      precision)
    return shape(geojson)

def convert_poly_coords(geom, affine_obj=None, inverse=False):
    """Georegister geometry objects currently in pixel coords or vice versa.

    Args:
        geom : :class:`shapely.geometry.shape` or str
            A :class:`shapely.geometry.shape`, or WKT string-formatted geometry
            object currently in pixel coordinates.
        
        affine_obj: list or :class:`affine.Affine`
            An affine transformation to apply to `geom` in the form of an
            ``[a, b, d, e, xoff, yoff]`` list or an :class:`affine.Affine` object.
            Required if not using `raster_src`.
        inverse : bool, optional
            If true, will perform the inverse affine transformation, going from
            geospatial coordinates to pixel coordinates.

    Returns:
        out_geom
            A geometry in the same format as the input with its coordinate system
            transformed to match the destination object.
    """

    if not affine_obj:
        raise ValueError("affine_obj must be provided.")
    
    if isinstance(affine_obj, Affine):
        affine_xform = affine_obj
    else:
        # assume it's a list in either gdal or "standard" order
        # (list_to_affine checks which it is)
        if len(affine_obj) == 9:  # if it's straight from rasterio
            affine_obj = affine_obj[0:6]
            affine_xform = Affine(*affine_obj)

    if inverse:  # geo->px transform
        affine_xform = ~affine_xform

    if isinstance(geom, str):
        # get the polygon out of the wkt string
        g = shapely.wkt.loads(geom)
    elif isinstance(geom, shapely.geometry.base.BaseGeometry):
        g = geom
    else:
        raise TypeError('The provided geometry is not an accepted format. '
                        'This function can only accept WKT strings and '
                        'shapely geometries.')

    xformed_g = shapely.affinity.affine_transform(g, [affine_xform.a,
                                                      affine_xform.b,
                                                      affine_xform.d,
                                                      affine_xform.e,
                                                      affine_xform.xoff,
                                                      affine_xform.yoff])
    if isinstance(geom, str):
        # restore to wkt string format
        xformed_g = shapely.wkt.dumps(xformed_g)

    return xformed_g

def affine_transform_gdf(gdf, affine_obj, inverse=False, geom_col="geometry", precision=None):
    """Perform an affine transformation on a GeoDataFrame.

    Args:
        gdf : :class:`geopandas.GeoDataFrame`, :class:`pandas.DataFrame`, or `str`
            A GeoDataFrame, pandas DataFrame with a ``"geometry"`` column (or a
            different column containing geometries, identified by `geom_col` -
            note that this column will be renamed ``"geometry"`` for ease of use
            with geopandas), or the path to a saved file in .geojson or .csv
            format.
        affine_obj : list or :class:`affine.Affine`
            An affine transformation to apply to `geom` in the form of an
            ``[a, b, d, e, xoff, yoff]`` list or an :class:`affine.Affine` object.
        inverse : bool, optional
            Use this argument to perform the inverse transformation.
        geom_col : str, optional
            The column in `gdf` corresponding to the geometry. Defaults to
            ``'geometry'``.
        precision : int, optional
            Decimal precision to round the geometries to. If not provided, no
            rounding is performed.
    Returns:
        gdf : :class:`geopandas.GeoDataFrame`
    """
    if isinstance(gdf, str):  # assume it's a geojson
        if gdf.lower().endswith('json'):
            gdf = gpd.read_file(gdf)
        elif gdf.lower().endswith('csv'):
            gdf = pd.read_csv(gdf)
        else:
            raise ValueError(
                "The file format is incompatible with this function.")
    if 'geometry' not in gdf.columns:
        gdf = gdf.rename(columns={geom_col: 'geometry'})
    if not isinstance(gdf['geometry'][0], Polygon):
        gdf['geometry'] = gdf['geometry'].apply(shapely.wkt.loads)
    gdf["geometry"] = gdf["geometry"].apply(convert_poly_coords,
                                            affine_obj=affine_obj,
                                            inverse=inverse)
    if precision is not None:
        gdf['geometry'] = gdf['geometry'].apply(
            _reduce_geom_precision, precision=precision)
    # the CRS is no longer valid - remove it
    gdf.crs = None

    return gdf

def geojson_to_px_gdf(geojson, im_path, geom_col='geometry',
                      output_path=None, override_crs=False):
    """Convert a geojson or set of geojsons from geo coords to px coords.

    Args:
        geojson : str
            Path to a geojson. This function will also accept a
            :class:`pandas.DataFrame` or :class:`geopandas.GeoDataFrame` with a
            column named ``'geometry'`` in this argument.
        im_path : str
            Path to a georeferenced image (ie a GeoTIFF) that geolocates to the
            same geography as the `geojson`(s). This function will also accept a
            :class:`osgeo.gdal.Dataset` or :class:`rasterio.DatasetReader` with
            georeferencing information in this argument.
        geom_col : str, optional
            The column containing geometry in `geojson`. If not provided, defaults
            to ``"geometry"``.
        output_path : str, optional
            Path to save the resulting output to. If not provided, the object
            won't be saved to disk.
        override_crs: bool, optional
            Useful if the geojsons generated by the vector tiler or otherwise were saved
            out with a non EPSG code projection. True sets the gdf crs to that of the 
            image, the inputs should have the same underlying projection for this to work. 
            If False, and the gdf does not have an EPSG code, this function will fail.
    Returns:
        output_df : :class:`pandas.DataFrame`
            A :class:`pandas.DataFrame` with all geometries in `geojson` that
            overlapped with the image at `im_path` converted to pixel coordinates.
            Additional columns are included with the filename of the source
            geojson (if available) and images for reference.

    """
    # get the bbox and affine transforms for the image
    im = rasterio_load(im_path)
    if isinstance(im_path, rasterio.DatasetReader):
        im_path = im_path.name
    # make sure the geo vector data is loaded in as geodataframe(s)
    gdf = gdf_load(geojson)

    if len(gdf):  # if there's at least one geometry
        if override_crs:
            gdf.crs = im.crs 
        overlap_gdf = get_overlapping_subset(gdf, im)
    else:
        overlap_gdf = gdf

    affine_obj = im.transform
    transformed_gdf = affine_transform_gdf(overlap_gdf, affine_obj=affine_obj,
                                           inverse=True, geom_col=geom_col)
    transformed_gdf['image_fname'] = os.path.split(im_path)[1]

    if output_path is not None:
        if output_path.lower().endswith('json'):
            transformed_gdf.to_file(output_path, driver='GeoJSON')
        else:
            transformed_gdf.to_csv(output_path, index=False)
    return transformed_gdf

def get_overlapping_subset(gdf, im=None, bbox=None, bbox_crs=None):
    """Extract a subset of geometries in a GeoDataFrame that overlap with `im`.

    Notes
    -----
    This function uses RTree's spatialindex, which is much faster (but slightly
    less accurate) than direct comparison of each object for overlap.

    Args:
        gdf : :class:`geopandas.GeoDataFrame`
            A :class:`geopandas.GeoDataFrame` instance or a path to a geojson.
        im : :class:`rasterio.DatasetReader` or `str`, optional
            An image object loaded with `rasterio` or a path to a georeferenced
            image (i.e. a GeoTIFF).
        bbox : `list` or :class:`shapely.geometry.Polygon`, optional
            A bounding box (either a :class:`shapely.geometry.Polygon` or a
            ``[bottom, left, top, right]`` `list`) from an image. Has no effect
            if `im` is provided (`bbox` is inferred from the image instead.) If
            `bbox` is passed and `im` is not, a `bbox_crs` should be provided to
            ensure correct geolocation - if it isn't, it will be assumed to have
            the same crs as `gdf`.
        bbox_crs : int, optional
            The coordinate reference system that the bounding box is in as an EPSG
            int. If not provided, it's assumed that the CRS is the same as `im`
            (if provided) or `gdf` (if not).

    Returns:
        output_gdf : :class:`geopandas.GeoDataFrame`
            A :class:`geopandas.GeoDataFrame` with all geometries in `gdf` that
            overlapped with the image at `im`.
            Coordinates are kept in the CRS of `gdf`.

    """
    if im is None and bbox is None:
        raise ValueError('Either `im` or `bbox` must be provided.')
    gdf = gdf_load(gdf)
    sindex = gdf.sindex
    if im is not None:
        im = rasterio_load(im)
        # currently, convert CRSs to WKT strings here to accommodate rasterio.
        bbox = transform_bounds(im.crs, check_crs(gdf.crs, return_rasterio=True),
                                *im.bounds)
        bbox_crs = im.crs
    # use transform_bounds in case the crs is different - no effect if not
    if isinstance(bbox, Polygon):
        bbox = bbox.bounds
    if bbox_crs is None:
        try:
            bbox_crs = check_crs(gdf.crs, return_rasterio=True)
        except AttributeError:
            raise ValueError('If `im` and `bbox_crs` are not provided, `gdf`'
                             'must provide a coordinate reference system.')
    else:
        bbox_crs = check_crs(bbox_crs, return_rasterio=True)
    # currently, convert CRSs to WKT strings here to accommodate rasterio.
    bbox = transform_bounds(bbox_crs, check_crs(gdf.crs, return_rasterio=True), *bbox)
    try:
        intersectors = list(sindex.intersection(bbox))
    except RTreeError:
        intersectors = []

    return gdf.iloc[intersectors, :]

def bbox_corners_to_coco(bbox):
    """Converts bbox from [minx, miny, maxx, maxy] to COCO format.

    Args:
        bbox (list-like of numerics): A 4-element list of the form [minx, miny, maxx, maxy].

    Returns:
        list: A list of the form [minx, miny, width, height].

    """
    return [bbox[0], bbox[1], bbox[2]-bbox[0], bbox[3]-bbox[1]]

def polygon_to_coco(polygon):
    """Converts a geometry to COCO polygon format.

    Args:
        polygon: A shapely geometry object or a Well-Known Text (WKT) string representing a polygon.

    Returns:
        A list of coordinates in COCO polygon format.

    Raises:
        ValueError: If the polygon is not a shapely geometry object or a WKT string.

    """
    if isinstance(polygon, Polygon):
        coords = polygon.exterior.coords.xy
    elif isinstance(polygon, str):  # assume it's WKT
        coords = loads(polygon).exterior.coords.xy
    else:
        raise ValueError('polygon must be a shapely geometry or WKT.')
    # zip together x,y pairs
    coords = list(zip(coords[0], coords[1]))
    coords = [item for coordinate in coords for item in coordinate]

    return coords

def coco_categories_dict_from_df(df, 
                                 category_id_col, 
                                 category_name_col, 
                                 supercategory_col=None):
    """Extracts category IDs, category names, and supercategory names from a DataFrame.

    Args:
        df (pandas.DataFrame): A DataFrame of records to filter for category info.
        category_id_col (str): The name for the column in `df` that contains category IDs.
        category_name_col (str): The name for the column in `df` that contains category names.
        supercategory_col (str, optional): The name for the column in `df` that contains supercategory names,
            if one exists. If not provided, supercategory will be left out of the output.

    Returns:
        list of dict: A list of dictionaries that contain category records per the COCO dataset specification.

    Raises:
        None

    Examples:
        >>> df = pd.DataFrame({'category_id': [1, 2, 3],
        ...                    'category_name': ['cat', 'dog', 'bird'],
        ...                    'supercategory': ['animal', 'animal', 'animal']})
        >>> coco_categories_dict_from_df(df, 'category_id', 'category_name', 'supercategory')
        [{'id': 1, 'name': 'cat', 'supercategory': 'animal'},
         {'id': 2, 'name': 'dog', 'supercategory': 'animal'},
         {'id': 3, 'name': 'bird', 'supercategory': 'animal'}]
    """
    cols_to_keep = [category_id_col, category_name_col]
    rename_dict = {category_id_col: 'id',
                   category_name_col: 'name'}
    if supercategory_col is not None:
        cols_to_keep.append(supercategory_col)
        rename_dict[supercategory_col] = 'supercategory'
    coco_cat_df = df[cols_to_keep]
    coco_cat_df = coco_cat_df.rename(columns=rename_dict)
    coco_cat_df = coco_cat_df.drop_duplicates()

    return coco_cat_df.to_dict(orient='records')

def _coco_category_name_id_dict_from_list(category_list):
    """Extracts a dictionary mapping category names to category IDs from a list.

    Args:
        category_list (list): A list of category dictionaries.

    Returns:
        dict: A dictionary mapping category names to category IDs.

    """
    # check if this is a full annotation json or just the categories
    category_dict = {category['name']: category['id']
                     for category in category_list}
    return category_dict

def make_coco_image_dict(image_ref, license_id=None):
    """Creates a COCO-formatted image record list from a dictionary of image filenames and IDs.

    Args:
        image_ref (dict): A dictionary of image filenames and their corresponding IDs.
        license_id (int, optional): The license ID number for the relevant license. Defaults to None.

    Returns:
        list: A list of COCO-formatted image records ready for export to JSON.
    """
    image_records = []
    for image_fname, image_id in image_ref.items():
        with rasterio.open(image_fname) as f:
            width = f.width
            height = f.height
        im_record = {'id': image_id,
                     'file_name': os.path.split(image_fname)[1],
                     'width': width,
                     'height': height}
        if license_id is not None:
            im_record['license'] = license_id
        image_records.append(im_record)

    return image_records

def df_to_coco_annos(df, output_path=None, geom_col='geometry',
                     image_id_col=None, category_col=None, score_col=None,
                     preset_categories=None, supercategory_col=None,
                     include_other=True, starting_id=1, verbose=0):
    """Extract COCO-formatted annotations from a pandas DataFrame.

    This function assumes that annotations are already in pixel coordinates.
    If this is not the case, you can transform them using
    solaris.vector.polygon.geojson_to_px_gdf.

    Note that this function generates annotations formatted per the COCO object
    detection specification. For additional information, see
    the COCO dataset specification.

    Args:
        df (pandas.DataFrame): A DataFrame containing geometries to store as annos.
        output_path (str, optional): The output file path to save the COCO-formatted annotations.
        geom_col (str, optional): The name of the column in df that contains geometries.
            The geometries should either be shapely.geometry.Polygon or WKT strings.
            Defaults to "geometry".
        image_id_col (str, optional): The column containing image IDs.
            If not provided, it's assumed that all are in the same image, which will be assigned the ID of 1.
        category_col (str, optional): The name of the column that specifies categories for each object.
            If not provided, all objects will be placed in a single category named "other".
        score_col (str, optional): The name of the column that specifies the output confidence of a model.
            If not provided, will not be output.
        preset_categories (list of dict, optional): A pre-set list of categories to use for labels.
            These categories should be formatted per the COCO category specification.
        supercategory_col (str, optional): The name of the column that specifies the supercategory for each category.
        include_other (bool, optional): Whether to include objects not contained in preset_categories.
            Defaults to True.
        starting_id (int, optional): The number to start numbering annotation IDs at. Defaults to 1.
        verbose (int, optional): Verbose text output. By default, none is provided;
            if True or 1, information-level outputs are provided; if 2, extremely verbose text is output.

    Returns:
        dict: A dictionary containing COCO-formatted annotation and category entries
            per the COCO dataset specification.
    """
    df = df_load(df)
    temp_df = df.copy()  # for manipulation
    if preset_categories is not None and category_col is None:
        logger.debug('preset_categories has a value, category_col is None.')
        raise ValueError('category_col must be specified if using'
                         ' preset_categories.')
    elif preset_categories is not None and category_col is not None:
        logger.debug('Both preset_categories and category_col have values.')
        logger.debug('Getting list of category names.')
        category_dict = _coco_category_name_id_dict_from_list(preset_categories)
        category_names = list(category_dict.keys())
        if not include_other:
            logger.debug('Filtering out objects not contained in '
                        ' preset_categories')
            temp_df = temp_df.loc[temp_df[category_col].isin(category_names),:]
        else:
            logger.debug('Setting category to "other" for objects outside of '
                        'preset category list.')
            temp_df.loc[~temp_df[category_col].isin(category_names),
                        category_col] = 'other'
            if 'other' not in category_dict.keys():
                logger.debug('Adding "other" to category_dict.')
                other_id = np.array(list(category_dict.values())).max() + 1
                category_dict['other'] = other_id
                preset_categories.append({'id': other_id,
                                          'name': 'other',
                                          'supercategory': 'other'})
    elif preset_categories is None and category_col is not None:
        logger.debug('No preset_categories, have category_col.')
        logger.debug(f'Collecting unique category names from {category_col}.')
        category_names = list(temp_df[category_col].unique())
        logger.debug('Generating category ID numbers arbitrarily.')
        category_dict = {k: v for k, v in zip(category_names,
                                              range(1, len(category_names)+1))}
    else:
        logger.debug('No category column or preset categories.')
        logger.debug('Setting category to "other" for all objects.')
        category_col = 'category_col'
        temp_df[category_col] = 'other'
        category_names = ['other']
        category_dict = {'other': 1}

    if image_id_col is None:
        temp_df['image_id'] = 1
    else:
        temp_df.rename(columns={image_id_col: 'image_id'})
    logger.debug('Checking geometries.')
    temp_df[geom_col] = temp_df[geom_col].apply(check_geom)
    logger.debug('Getting area of geometries.')
    temp_df['area'] = temp_df[geom_col].apply(lambda x: x.area)
    logger.debug('Getting geometry bounding boxes.')
    temp_df['bbox'] = temp_df[geom_col].apply(lambda x: bbox_corners_to_coco(x.bounds))
    temp_df['category_id'] = temp_df[category_col].map(category_dict)
    temp_df['annotation_id'] = list(range(starting_id,
                                          starting_id + len(temp_df)))
    if score_col is not None:
        temp_df['score'] = df[score_col]

    def _row_to_coco(row, geom_col, category_id_col, image_id_col, score_col):
        "get a single annotation record from a row of temp_df."
        if score_col is None:

            return {'id': row['annotation_id'],
                    'image_id': int(row[image_id_col]),
                    'category_id': int(row[category_id_col]),
                    'segmentation': [polygon_to_coco(row[geom_col])],
                    'area': row['area'],
                    'bbox': row['bbox'],
                    'iscrowd': 0}
        else:
            return {'id': row['annotation_id'],
                    'image_id': int(row[image_id_col]),
                    'category_id': int(row[category_id_col]),
                    'segmentation': [polygon_to_coco(row[geom_col])],
                    'score': float(row[score_col]),
                    'area': row['area'],
                    'bbox': row['bbox'],
                    'iscrowd': 0}

    coco_annotations = temp_df.apply(_row_to_coco, axis=1, geom_col=geom_col,
                                     category_id_col='category_id',
                                     image_id_col=image_id_col,
                                     score_col=score_col).tolist()
    coco_categories = coco_categories_dict_from_df(
        temp_df, category_id_col='category_id',
        category_name_col=category_col,
        supercategory_col=supercategory_col)

    output_dict = {'annotations': coco_annotations,
                   'categories': coco_categories}

    if output_path is not None:
        with open(output_path, 'w') as outfile:
            json.dump(output_dict, outfile)

    return output_dict



import argparse
import logging
import os
import re
import tarfile
from pathlib import Path
from urllib.parse import urlparse

import requests
import torch
import yaml

from ..config.logging_config import logger

logger = logging.getLogger(__name__)

USER_CACHE = Path.home().joinpath(".cache")
script_dir = Path(__file__).resolve().parent.parent
MODEL_CONFIG = script_dir / "config" /  "models.yaml"

def is_tiff_path(path: str):
    # Check if the given path ends with .tiff or .tif (case insensitive)
    return re.match(r'.*\.(tiff|tif)$', path, re.IGNORECASE) is not None

def is_tiff_url(url: str):
    # Check if the URL ends with .tiff or .tif (case insensitive)
    parsed_url = urlparse(url)
    return re.match(r'.*\.(tiff|tif)$', os.path.basename(parsed_url.path), re.IGNORECASE) is not None

def read_yaml(yaml_file_path: str or Path):
    with open(yaml_file_path, "r") as f:
        config = yaml.safe_load(f.read())
    return config

def validate_asset_type(image_asset: str):
    """Validate image asset type

    Args:
        image_asset (str): File path or URL to the image asset.

    Returns:
        str: file path or url.
    """
    if os.path.isfile(image_asset) and is_tiff_path(image_asset):
        return image_asset
    elif urlparse(image_asset).scheme in ('http', 'https') and is_tiff_url(image_asset):
        return image_asset
    return None

def calculate_gpu_stats(gpu_id: int = 0):
    """Calculate GPU stats

    Args:
        gpu_id (int, optional): GPU id. Defaults to 0.

    Returns:
        tuple(dict, dict): gpu stats.
    """
    res = {'gpu': torch.cuda.utilization(gpu_id)}
    torch_cuda_mem = torch.cuda.mem_get_info(gpu_id)
    mem = {
        'used': torch_cuda_mem[-1] - torch_cuda_mem[0],
        'total': torch_cuda_mem[-1]
    }    
    return res, mem

def download_file_from_url(url, save_path, access_token=None):
    """Download a file from a URL
    
    Args:
        url (str): URL to the file.
        save_path (str or Path): Path to save the file.
        access_token (str, optional): Access token. Defaults to None.
    """
    try:
        headers = {}
        headers["Authorization"] = f"Bearer {access_token}"
        response = requests.get(url, headers=headers, stream=True)
        if response.status_code == 200:
            with open(save_path, 'wb') as file:
                for chunk in response.iter_content(chunk_size=128):
                    file.write(chunk)
            logger.info(f"Downloaded {save_path}")
        else:
            logger.error(f"Failed to download the file from {url}. Status code: {response.status_code}")
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        raise

def extract_tar_gz(tar_gz_file: str or Path, target_directory: str or Path):
    """Extracts a tar.gz file to a target directory
    Args:
        tar_gz_file (str or Path): Path to the tar.gz file.
        target_directory (str o rPath): Path to the target directory.
    """
    try:
        with tarfile.open(tar_gz_file, "r:gz") as tar:
            for member in tar.getmembers():
                if member.isreg():
                    member.name = os.path.basename(member.name)
                    tar.extract(member, target_directory) 
            # tar.extractall(path=target_directory)
        logger.info(f"Successfully extracted {tar_gz_file} to {target_directory}")
        os.remove(tar_gz_file)
    except tarfile.TarError as e:
        logger.error(f"Error while extracting {tar_gz_file}: {e}")
    except Exception as e:
        logger.error(f"An error occurred: {e}") 
        raise

def get_device(device: str = "gpu",
               gpu_id: int = 0,
               gpu_max_ram_usage: int = 25,
               gpu_max_utilization: int = 15):
    """Returns a torch device

    Args:
        device (str): Accepts "cpu" or "gpu". Defaults to "gpu".
        gpu_id (int): GPU id. Defaults to 0.
        gpu_max_ram_usage (int): GPU max ram usage. Defaults to 25.
        gpu_max_utilization (int): GPU max utilization. Defaults to 15.

    Returns:
        torch.device: torch device.
    """
    if device == "cpu":
        return torch.device('cpu')
    elif device == "gpu":
        res, mem = calculate_gpu_stats(gpu_id=gpu_id)
        used_ram = mem['used'] / (1024 ** 2)
        max_ram = mem['total'] / (1024 ** 2)
        used_ram_percentage = (used_ram / max_ram) * 100
        logger.info(f"\nGPU RAM used: {round(used_ram_percentage)}%" 
                    f"[used_ram: {used_ram:.0f}MiB] [max_ram: {max_ram:.0f}MiB]\n"
                    f"GPU Utilization: {res['gpu']}%")
        if used_ram_percentage < gpu_max_ram_usage:
            if res["gpu"] < gpu_max_utilization:
                return torch.device(f"cuda:{gpu_id}")
            else:
                logger.warning(f"Reverting to CPU!\n"
                              f"Current GPU:{gpu_id} utilization: {res['gpu']}%\n"
                              f"Max GPU utilization allowed: {gpu_max_utilization}%")
                return torch.device('cpu')
        else:
            logger.warning(f"Reverting to CPU!\n"
                           f"Current GPU:{gpu_id} RAM usage: {used_ram_percentage}%\n"
                           f"Max used RAM allowed: {gpu_max_ram_usage}%")
            return torch.device('cpu')
    else:
        logger.error("Invalid device type requested: {device}")
        raise ValueError("Invalid device type")

def get_directory(work_directory: str)-> Path:
    """Returns a working directory

    Args:
        work_directory (str): User's specified path

    Returns:
        Path: working directory
    """
    
    if work_directory:
        work_directory = Path(work_directory)
        if not work_directory.is_dir():
            Path.mkdir(work_directory, parents=True)
    else:
        work_directory = USER_CACHE.joinpath("geo-inference")
        if not work_directory.is_dir():
            Path.mkdir(work_directory, parents=True)
    
    return work_directory

def get_model(model_path_or_url: str, work_dir: Path) -> Path:
    """Download a model from the model zoo

    Args:
        model_path_or_url (str): Path or url of model file.
        work_dir (Path): Working directory.

    Returns:
        Path: Path to the model file.
    """
    parsed_string = urlparse(model_path_or_url)
    if parsed_string.scheme:
        model_url = model_path_or_url
        model_name = os.path.basename(parsed_string.path)
        cached_file = work_dir.joinpath(model_name)
        if not cached_file.is_file():
            download_file_from_url(model_url, cached_file)
        return cached_file
    else:
        model_path = Path(model_path_or_url)
        if model_path.is_file():
            return model_path
        else:
            logger.error(f"Model {model_path_or_url} not found")
            raise ValueError("Invalid model path")

def cmd_interface(argv=None):
    """
    Parse command line arguments for extracting features from high-resolution imagery using pre-trained models.

    Args:
        argv (list): List of arguments to parse. If None, the arguments are taken from sys.argv.

    Returns:
        dict: A dictionary containing the parsed arguments.

    Raises:
        SystemExit: If the arguments are not valid.

    Usage:
        Use the -h option to get supported arguments.
    """
    parser = argparse.ArgumentParser(usage="%(prog)s [-h HELP] use -h to get supported arguments.",
                                     description='Extract features from high-resolution imagery using pre-trained models.')
    
    parser.add_argument("-a", "--args", nargs=1, help="Path to arguments stored in yaml, consult ./config/sample_config.yaml")
    
    parser.add_argument("-i", "--image", nargs=1, help="Path to Geotiff")
    
    parser.add_argument("-bb", "--bbox", nargs=1, help="AOI bbox in this format'minx, miny, maxx, maxy'")
    
    parser.add_argument("-m", "--model", nargs=1, help="Path or URL to the model file")
    
    parser.add_argument("-wd", "--work_dir", nargs=1, help="Working Directory")
    
    parser.add_argument("-bs", "--batch_size", nargs=1, help="The Batch Size")
    
    parser.add_argument("-v", "--vec", nargs=1, help="Vector Conversion")
    
    parser.add_argument("-d", "--device", nargs=1, help="CPU or GPU Device")
    
    parser.add_argument("-id", "--gpu_id", nargs=1, help="GPU ID, Default = 0")
    
    args = parser.parse_args()
    
    if args.args:
        config = read_yaml(args.args[0])
        image = config["arguments"]["image"]
        bbox = config["arguments"]["bbox"]
        model = config["arguments"]["model"]
        work_dir = config["arguments"]["work_dir"]
        batch_size = config["arguments"]["batch_size"]
        vec = config["arguments"]["vec"]
        device = config["arguments"]["device"]
        gpu_id = config["arguments"]["gpu_id"]
    elif args.image:
        image = args.image[0]
        bbox = args.bbox[0] if args.bbox else None
        model = args.model[0] if args.model else None
        work_dir = args.work_dir[0] if args.work_dir else None
        batch_size = args.batch_size[0] if args.batch_size else 1
        vec = args.vec[0] if args.vec else False
        device = args.device[0] if args.device else "gpu"
        gpu_id = args.gpu_id[0] if args.gpu_id else 0
    else:
        print('use the help [-h] option for correct usage')
        raise SystemExit
    
    arguments= {"image": image,
                "bbox": bbox,
                "model": model,
                "work_dir": work_dir,
                "batch_size": batch_size,
                "vec": vec,
                "device": device,
                "gpu_id": gpu_id
                }
    return arguments
    
if __name__ == '__main__':
    pass
    


import logging
import time
from pathlib import Path

import torch
from torch.utils.data import DataLoader
from torchgeo.datasets import stack_samples
from tqdm import tqdm

from .config.logging_config import logger
from .geo_blocks import InferenceMerge, InferenceSampler, RasterDataset
from .utils.helpers import cmd_interface, get_device, get_directory, get_model
from .utils.polygon import gdf_to_yolo, mask_to_poly_geojson, geojson2coco

logger = logging.getLogger(__name__)


class GeoInference:
    """
    A class for performing geo inference on geospatial imagery using a pre-trained model.

    Args:
        model (str): The path or url to the model file
        work_dir (str): The directory where the model and output files will be saved.
        batch_size (int): The batch size to use for inference.
        mask_to_vec (bool): Whether to convert the output mask to vector format.
        device (str): The device to use for inference (either "cpu" or "gpu").
        gpu_id (int): The ID of the GPU to use for inference (if device is "gpu").

    Attributes:
        batch_size (int): The batch size to use for inference.
        work_dir (Path): The directory where the model and output files will be saved.
        device (torch.device): The device to use for inference.
        mask_to_vec (bool): Whether to convert the output mask to vector format.
        model (torch.jit.ScriptModule): The pre-trained model to use for inference.
        classes (int): The number of classes in the output of the model.

    """

    def __init__(self,
                 model: str = None,
                 work_dir: str = None,
                 batch_size: int = 1,
                 mask_to_vec: bool = False,
                 device: str = "gpu",
                 gpu_id: int = 0):
        self.gpu_id = int(gpu_id)
        self.batch_size = int(batch_size)
        self.work_dir: Path = get_directory(work_dir)
        self.device = get_device(device=device, 
                                 gpu_id=self.gpu_id)
        model_path: Path = get_model(model_path_or_url=model, 
                                     work_dir=self.work_dir)
        self.mask_to_vec = mask_to_vec
        self.model = torch.jit.load(model_path, map_location=self.device)
        dummy_input = torch.ones((1, 3, 32, 32), device=self.device, dtype=torch.float)
        with torch.no_grad():
            self.classes = self.model(dummy_input).shape[1]
    
    @torch.no_grad() 
    def __call__(self, tiff_image: str, bbox: str = None, patch_size: int = 512, stride_size: str = None) -> None:
        """
        Perform geo inference on geospatial imagery.

        Args:
            tiff_image (str): The path to the geospatial image to perform inference on.
            bbox (str): The bbox or extent of the image in this format "minx, miny, maxx, maxy"
            patch_size (int): The size of the patches to use for inference.
            stride_size (int): The stride to use between patches.

        Returns:
            None

        """
        mask_path = self.work_dir.joinpath(Path(tiff_image).stem + "_mask.tif")
        polygons_path = self.work_dir.joinpath(Path(tiff_image).stem + "_polygons.geojson")
        yolo_csv_path = self.work_dir.joinpath(Path(tiff_image).stem + "_yolo.csv")
        coco_json_path = self.work_dir.joinpath(Path(tiff_image).stem + "_coco.json")
        
        dataset = RasterDataset(tiff_image, bbox=bbox)
        sampler = InferenceSampler(dataset, size=patch_size, stride=patch_size >> 1 if stride_size is None else stride_size, roi=dataset.bbox)
        roi_height = sampler.im_height 
        roi_width = sampler.im_width
        h_padded, w_padded = roi_height + patch_size, roi_width + patch_size
        output_meta = dataset.src.meta
        merge_patches = InferenceMerge(height=h_padded, width=w_padded, classes=self.classes, device=self.device)
        dataloader = DataLoader(dataset, batch_size=self.batch_size, sampler=sampler, collate_fn=stack_samples)
        
        start_time = time.time()
        
        for batch in tqdm(dataloader, desc='extracting features', unit='batch', total=len(dataloader)):
            image_tensor = batch["image"].to(self.device)
            window_tensor = batch["window"].unsqueeze(1).to(self.device)
            pixel_xy = batch["pixel_coords"]
            output = self.model(image_tensor) 
            merge_patches.merge_on_cpu(batch=output, windows=window_tensor, pixel_coords=pixel_xy)
        merge_patches.save_as_tiff(height=roi_height, 
                                   width=roi_width, 
                                   output_meta=output_meta, 
                                   output_path=mask_path)
        
        if self.mask_to_vec:
            mask_to_poly_geojson(mask_path, polygons_path)
            gdf_to_yolo(polygons_path, mask_path, yolo_csv_path)
            geojson2coco(mask_path, polygons_path, coco_json_path)
            
        end_time = time.time() - start_time
        
        logger.info('Extraction Completed in {:.0f}m {:.0f}s'.format(end_time // 60, end_time % 60))

def main() -> None:
    arguments = cmd_interface()
    geo_inference = GeoInference(model=arguments["model"],
                                 work_dir=arguments["work_dir"],
                                 batch_size=arguments["batch_size"],
                                 mask_to_vec=arguments["vec"],
                                 device=arguments["device"],
                                 gpu_id=arguments["gpu_id"])
    geo_inference(tiff_image=arguments["image"], bbox=arguments["bbox"])
               
if __name__ == "__main__":
    main()


import os
import pytest
import torch
from geo_inference.geo_inference import GeoInference
from pathlib import Path

@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "data"

class TestGeoInference:
    @pytest.fixture
    def geo_inference(self, test_data_dir):
        model = str(test_data_dir / "inference"/ "test_model" / "test_model.pt")
        work_dir = str(test_data_dir / "inference")
        batch_size = 1
        mask_to_vec = True
        device = 'cpu'
        gpu_id = 0
        return GeoInference(model, work_dir, batch_size, mask_to_vec, device, gpu_id)

    def test_init(self, geo_inference, test_data_dir):
        assert geo_inference.gpu_id == 0
        assert geo_inference.batch_size == 1
        assert geo_inference.work_dir == test_data_dir / "inference"
        assert geo_inference.device == torch.device('cpu')
        assert geo_inference.mask_to_vec == True
        assert isinstance(geo_inference.model, torch.jit.ScriptModule)
        assert geo_inference.classes > 0

    def test_call(self, geo_inference, test_data_dir):
        tiff_image = test_data_dir / '0.tif'
        # bbox = '0,0,100,100'
        # patch_size = 512
        # stride_size = 256
        geo_inference(str(tiff_image))
        mask_path = geo_inference.work_dir / "0_mask.tif"
        assert mask_path.exists()
        if geo_inference.mask_to_vec:
            polygons_path = geo_inference.work_dir / "0_polygons.geojson"
            yolo_csv_path = geo_inference.work_dir / "0_yolo.csv"
            coco_path = geo_inference.work_dir / "0_coco.json"
            assert polygons_path.exists()
            assert yolo_csv_path.exists()
            assert coco_path.exists()
            os.remove(polygons_path)
            os.remove(yolo_csv_path)
            os.remove(coco_path)
        os.remove(mask_path)